package org.opencdmp.filetransformer.rda.service.rdafiletransformer;

import gr.cite.tools.exception.MyApplicationException;
import gr.cite.tools.logging.LoggerService;
import org.opencdmp.commonmodels.enums.ContactInfoType;
import org.opencdmp.commonmodels.enums.DmpUserRole;
import org.opencdmp.commonmodels.models.*;
import org.opencdmp.commonmodels.models.description.DescriptionModel;
import org.opencdmp.commonmodels.models.description.PropertyDefinitionFieldSetItemModel;
import org.opencdmp.commonmodels.models.description.PropertyDefinitionFieldSetModel;
import org.opencdmp.commonmodels.models.description.PropertyDefinitionModel;
import org.opencdmp.commonmodels.models.descriptiotemplate.FieldModel;
import org.opencdmp.commonmodels.models.descriptiotemplate.FieldSetModel;
import org.opencdmp.commonmodels.models.dmp.DmpBlueprintValueModel;
import org.opencdmp.commonmodels.models.dmp.DmpModel;
import org.opencdmp.commonmodels.models.dmpblueprint.SectionModel;
import org.opencdmp.commonmodels.models.dmpreference.DmpReferenceModel;
import org.opencdmp.commonmodels.models.reference.ReferenceFieldModel;
import org.opencdmp.commonmodels.models.reference.ReferenceModel;
import org.opencdmp.filetransformer.rda.model.rda.*;
import org.opencdmp.filetransformer.rda.service.descriptiontemplatesearcher.TemplateFieldSearcherService;
import org.opencdmp.filetransformerbase.enums.FileTransformerEntityType;
import org.opencdmp.filetransformerbase.interfaces.FileTransformerClient;
import org.opencdmp.filetransformerbase.interfaces.FileTransformerConfiguration;
import org.opencdmp.filetransformer.rda.service.json.JsonHandlingService;
import org.opencdmp.filetransformer.rda.service.storage.FileStorageService;
import org.opencdmp.filetransformerbase.models.misc.FileFormat;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.ConfigurableBeanFactory;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Service;

import java.net.URI;
import java.nio.charset.StandardCharsets;
import java.time.Instant;
import java.util.*;

@Service
@Scope(value = ConfigurableBeanFactory.SCOPE_PROTOTYPE)
public class RdaFileTransformerService implements FileTransformerClient {
    private static final LoggerService logger = new LoggerService(LoggerFactory.getLogger(RdaFileTransformerService.class));

    private final FileStorageService storageService;
    private final JsonHandlingService jsonHandlingService;
    private final RdaFileTransformerServiceConfiguration configuration;
    private final TemplateFieldSearcherService templateFieldSearcherService;
    
    private final static List<FileTransformerEntityType> FILE_TRANSFORMER_ENTITY_TYPES = List.of(FileTransformerEntityType.Dmp);
    private final static String REPOSITORY_ZENODO = "Zenodo";
    private final static String PREFIX_ORCID = "orcid";
    private final static String ORCID_URL = "http://orcid.org/";
    private static final String SEMANTIC_ETHICAL_ISSUES_EXISTS = "rda.dmp.ethical_issues_exist";
    private static final String SEMANTIC_ETHICAL_ISSUES_DESCRIPTION = "rda.dmp.ethical_issues_description";
    private static final String SEMANTIC_ETHICAL_ISSUES_REPORT = "rda.dmp.ethical_issues_report";
    private static final String SEMANTIC_COST_CURRENCY_CODE = "rda.dmp.cost.currency_code";
    private static final String SEMANTIC_COST_DESCRIPTION = "rda.dmp.cost.description";
    private static final String SEMANTIC_COST_TITLE = "rda.dmp.cost.title";
    private static final String SEMANTIC_COST_VALUE = "rda.dmp.cost.value";
    private static final String SEMANTIC_DATASET_LANGUAGE = "rda.dataset.language";
    private static final String SEMANTIC_DATASET_TYPE = "rda.dataset.type";
    private static final String SEMANTIC_DATASET_ISSUED = "rda.dataset.issued";
    private static final String SEMANTIC_DATASET_KEYWORD = "rda.dataset.keyword";
    private static final String SEMANTIC_DATASET_PERSONAL_DATA = "rda.dataset.personal_data";
    private static final String SEMANTIC_DATASET_SENSITIVE_DATA = "rda.dataset.sensitive_data";
    private static final String SEMANTIC_DATASET_PRESERVATION_STATEMENT = "rda.dataset.preservation_statement";
    private static final String SEMANTIC_DATASET_DATA_QUALITY_ASSURANCE = "rda.dataset.data_quality_assurance";
    private static final String SEMANTIC_DATASET_TECHNICAL_RESOURCE_DESCRIPTION = "rda.dataset.technical_resource.description";
    private static final String SEMANTIC_DATASET_TECHNICAL_RESOURCE_NAME = "rda.dataset.technical_resource.name";
    private static final String SEMANTIC_DATASET_SECURITY_AND_PRIVACY_DESCRIPTION = "rda.dataset.security_and_privacy.description";
    private static final String SEMANTIC_DATASET_SECURITY_AND_PRIVACY_TITLE = "rda.dataset.security_and_privacy.title";
    private static final String SEMANTIC_DATASET_METADATA_LANGUAGE = "rda.dataset.metadata.language";
    private static final String SEMANTIC_DATASET_METADATA_DESCRIPTION = "rda.dataset.metadata.description";
    private static final String SEMANTIC_DATASET_METADATA_STANDARD_ID = "rda.dataset.metadata.metadata_standard_id";
    private static final String SEMANTIC_DATASET_METADATA_STANDARD_ID_IDENTIFIER = "rda.dataset.metadata.metadata_standard_id.identifier";
    private static final String SEMANTIC_DATASET_METADATA_STANDARD_ID_TYPE = "rda.dataset.metadata.metadata_standard_id.type";
    private static final String SEMANTIC_DATASET_DATASET_ID = "rda.dataset.dataset_id";
    private static final String SEMANTIC_DATASET_DATASET_ID_ID = "rda.dataset.dataset_id.identifier";
    private static final String SEMANTIC_DATASET_DATASET_ID_TYPE = "rda.dataset.dataset_id.type";
    private static final String SEMANTIC_DATASET_DISTRIBUTION_ACCESS_URL = "rda.dataset.distribution.access_url";
    private static final String SEMANTIC_DATASET_DISTRIBUTION_AVAILABLE_UTIL = "rda.dataset.distribution.available_until";
    private static final String SEMANTIC_DATASET_DISTRIBUTION_BYTE_SIZE = "rda.dataset.distribution.byte_size";
    private static final String SEMANTIC_DATASET_DISTRIBUTION_DATA_ACCESS = "rda.dataset.distribution.data_access";
    private static final String SEMANTIC_DATASET_DISTRIBUTION_DESCRIPTION = "rda.dataset.distribution.description";
    private static final String SEMANTIC_DATASET_DISTRIBUTION_DOWNLOAD_URL = "rda.dataset.distribution.download_url";
    private static final String SEMANTIC_DATASET_DISTRIBUTION_FORMAT = "rda.dataset.distribution.format";
    private static final String SEMANTIC_DATASET_DISTRIBUTION_TITLE = "rda.dataset.distribution.title";
    private static final String SEMANTIC_DATASET_DISTRIBUTION_LICENCE = "rda.dataset.distribution.license";
    private static final String SEMANTIC_DATASET_DISTRIBUTION_HOST_AVAILABILITY = "rda.dataset.distribution.host.availability";
    private static final String SEMANTIC_DATASET_DISTRIBUTION_HOST_BACKUP_FREQUENCY = "rda.dataset.distribution.host.backup_frequency";
    private static final String SEMANTIC_DATASET_DISTRIBUTION_HOST_BACKUP_TYPE = "rda.dataset.distribution.host.backup_type";
    private static final String SEMANTIC_DATASET_DISTRIBUTION_HOST_CERTIFIED_WITH = "rda.dataset.distribution.host.certified_with";
    private static final String SEMANTIC_DATASET_DISTRIBUTION_HOST_DESCRIPTION = "rda.dataset.distribution.host.description";
    private static final String SEMANTIC_DATASET_DISTRIBUTION_HOST_GEO_LOCATION = "rda.dataset.distribution.host.geo_location";
    private static final String SEMANTIC_DATASET_DISTRIBUTION_HOST_PID_SYSTEM = "rda.dataset.distribution.host.pid_system";
    private static final String SEMANTIC_DATASET_DISTRIBUTION_HOST_STORAGE_TYPE = "rda.dataset.distribution.host.storage_type";
    private static final String SEMANTIC_DATASET_DISTRIBUTION_HOST_SUPPORT_VERSIONING = "rda.dataset.distribution.host.support_versioning";
    private static final String SEMANTIC_DATASET_DISTRIBUTION_HOST_TITLE = "rda.dataset.distribution.host.title";
    private static final String SEMANTIC_DATASET_DISTRIBUTION_HOST_URL = "rda.dataset.distribution.host.url";
    private static final String SEMANTIC_DATASET_CONTRIBUTOR = "rda.dmp.contributor";
    
    @Autowired
    public RdaFileTransformerService(FileStorageService storageService, JsonHandlingService jsonHandlingService, RdaFileTransformerServiceConfiguration configuration, TemplateFieldSearcherService templateFieldSearcherService) {
        this.storageService = storageService;
	    this.configuration = configuration;
	    this.jsonHandlingService = jsonHandlingService;
	    this.templateFieldSearcherService = templateFieldSearcherService;
    }
    @Override
    public FileEnvelopeModel exportDmp(DmpModel dmpFileTransformerModel, String variant)  {
        Dmp dmp = this.buildRdaDmp(dmpFileTransformerModel);
        RDAModel rdaModel = new RDAModel();
        rdaModel.setDmp(dmp);
        String dmpJson = jsonHandlingService.toJsonSafe(rdaModel);
        
        byte[] bytes = dmpJson.getBytes(StandardCharsets.UTF_8);
        FileEnvelopeModel wordFile = new FileEnvelopeModel();
        if (this.getConfiguration().isUseSharedStorage()) {
            String fileRef = this.storageService.storeFile(bytes);
            wordFile.setFileRef(fileRef);
        } else {
            wordFile.setFile(bytes);
        }
        wordFile.setFilename(dmpFileTransformerModel.getLabel() + ".json");
        return wordFile;
    }

    @Override
    public FileEnvelopeModel exportDescription(DescriptionModel descriptionFileTransformerModel, String format) {
        throw new MyApplicationException("Not implemented");
//        Dmp dmp = this.dmpRDAMapper.toRDA(descriptionFileTransformerModel.getDmp());
//        Map<String, Object> datasetExtraData = new HashMap<>();
//        datasetExtraData.put("dmp", dmp);
//        Dataset dataset = this.descriptionRDAMapper.toRDA(descriptionFileTransformerModel, datasetExtraData);
//        String dmpJson = jsonHandlingService.toJsonSafe(dataset);
//
//        byte[] bytes = dmpJson.getBytes(StandardCharsets.UTF_8);
//        FileEnvelopeModel wordFile = new FileEnvelopeModel();
//        if (this.getConfiguration().isUseSharedStorage()) {
//            String fileRef = this.storageService.storeFile(bytes);
//            wordFile.setFileRef(fileRef);
//        } else {
//            wordFile.setFile(bytes);
//        }
//        wordFile.setFilename(descriptionFileTransformerModel.getLabel() + ".json");
//        return wordFile;
    }
    
    //region export
    
    //region dmp
    
    private Dmp buildRdaDmp(DmpModel dmp){
        if (dmp == null) throw new MyApplicationException("DMP is missing");
        if (dmp.getCreatedAt() == null) throw new MyApplicationException("DMP Created is missing");
        if (dmp.getUpdatedAt() == null) throw new MyApplicationException("DMP Modified is missing");
        if (dmp.getLabel() == null) throw new MyApplicationException("DMP Label is missing");
        if (dmp.getProperties() == null)  throw new MyApplicationException("DMP is missing language and contact properties");

        List<ReferenceModel> grants = this.getReferenceModelOfTypeCode(dmp, this.configuration.getRdaFileTransformerServiceProperties().getGrantReferenceCode());
        List<ReferenceModel> researchers = this.getReferenceModelOfTypeCode(dmp, this.configuration.getRdaFileTransformerServiceProperties().getResearcherReferenceCode());
        List<ReferenceModel> funders = this.getReferenceModelOfTypeCode(dmp, this.configuration.getRdaFileTransformerServiceProperties().getFunderReferenceCode());
        List<ReferenceModel> projects = this.getReferenceModelOfTypeCode(dmp, this.configuration.getRdaFileTransformerServiceProperties().getProjectReferenceCode());
        List<ReferenceModel> licences = this.getReferenceModelOfTypeCode(dmp, this.configuration.getRdaFileTransformerServiceProperties().getLicenceReferenceCode());

        Dmp rda = new Dmp();
        rda.setDmpId(this.buildRdaDmpId(dmp));
        rda.setCreated(dmp.getCreatedAt());
        rda.setDescription(dmp.getDescription());
        rda.setModified(dmp.getUpdatedAt());
        rda.setTitle(dmp.getLabel());
        rda.setLanguage(this.buildRdaLanguage(dmp.getLanguage() != null ? dmp.getLanguage() : "en"));
        rda.setContributor(buildRdaContributors(researchers));
        rda.setContributor(buildRdaContributorsByUsers(dmp.getUsers()));
        rda.setContact(buildRdaContact(dmp));
        rda.setEthicalIssuesExist(buildRdaEthicalIssuesExist(dmp));
        rda.setEthicalIssuesDescription(buildRdaEthicalIssuesDescription(dmp));
        rda.setEthicalIssuesReport(buildRdaEthicalIssueReport(dmp));
        if (projects.size() == 1 && grants.size() == 1 && funders.size() == 1) { //TODO: Project hierarchy not implemented should rewrite this logic 
            rda.setProject(buildRdaProjects(projects, grants, funders));
        } else if (!projects.isEmpty()){
            rda.setProject(buildRdaProjects(projects, new ArrayList<>(), new ArrayList<>()));
        } else if (grants.size() == 1 && funders.size() == 1) {
            rda.setProject(buildRdaProjects(grants, grants, funders));
        }
        Cost cost = buildRdaCost(dmp);
        rda.setCost(new ArrayList<>());
        if (cost != null) rda.getCost().add(cost);

        if (dmp.getDescriptions() != null){
            rda.setDataset(new ArrayList<>());
            for (DescriptionModel descriptionModel : dmp.getDescriptions()) rda.getDataset().add(this.buildRdaDataset(descriptionModel, rda));
        }
        
        return rda;
    }

    public Cost buildRdaCost(DmpModel model) {
        if (model == null) throw new MyApplicationException("Dmp is missing");
        Cost cost = new Cost();
        
        List<org.opencdmp.commonmodels.models.dmpblueprint.FieldModel> titleFields = this.getFieldsOfSemantic(model, SEMANTIC_COST_TITLE);
        for(org.opencdmp.commonmodels.models.dmpblueprint.FieldModel field : titleFields) {
            DmpBlueprintValueModel valueModel = field != null ? this.getDmpBlueprintValue(model, field.getId()) : null;
            if (valueModel != null && valueModel.getValue() != null && !valueModel.getValue().isBlank()) cost.setTitle(valueModel.getValue());
        }

        List<org.opencdmp.commonmodels.models.dmpblueprint.FieldModel> descriptionFields = this.getFieldsOfSemantic(model, SEMANTIC_COST_DESCRIPTION);
        for(org.opencdmp.commonmodels.models.dmpblueprint.FieldModel field : descriptionFields) {
            DmpBlueprintValueModel valueModel = field != null ? this.getDmpBlueprintValue(model, field.getId()) : null;
            if (valueModel != null && valueModel.getValue() != null && !valueModel.getValue().isBlank()) cost.setDescription(valueModel.getValue());
        }

        List<org.opencdmp.commonmodels.models.dmpblueprint.FieldModel> valueFields = this.getFieldsOfSemantic(model, SEMANTIC_COST_VALUE);
        for(org.opencdmp.commonmodels.models.dmpblueprint.FieldModel field : valueFields) {
            DmpBlueprintValueModel valueModel = field != null ? this.getDmpBlueprintValue(model, field.getId()) : null;
            if (valueModel != null && valueModel.getNumberValue() != null) cost.setValue(valueModel.getNumberValue());
        }

        List<org.opencdmp.commonmodels.models.dmpblueprint.FieldModel> codeFields = this.getFieldsOfSemantic(model, SEMANTIC_COST_CURRENCY_CODE);
        for(org.opencdmp.commonmodels.models.dmpblueprint.FieldModel field : codeFields) {
            DmpBlueprintValueModel valueModel = field != null ? this.getDmpBlueprintValue(model, field.getId()) : null;
            if (valueModel != null && valueModel.getValue() != null && !valueModel.getValue().isBlank()) {
                try {
                    cost.setCurrencyCode(Cost.CurrencyCode.fromValue(valueModel.getValue()));
                } catch (Exception e) {
                    logger.error(e.getMessage(), e);
                    logger.warn("Dmp cost currency code is not valid and will not be set.");
                }
            }
        }
        return cost.getValue() != null ? cost : null;
    }

    public Project buildRdaProject(ReferenceModel project, List<ReferenceModel> grants,  List<ReferenceModel> funders) {
        if (project == null) throw new MyApplicationException("Project is missing");
        if (project.getLabel() == null || project.getLabel().isBlank()) throw new MyApplicationException("Project Title is missing");
        Project rda = new Project();
        rda.setTitle(project.getLabel());
        rda.setDescription(project.getDescription());
        if (project.getDefinition() != null && project.getDefinition().getFields() != null) {
	        project.getDefinition().getFields().stream().filter(field -> this.configuration.getRdaFileTransformerServiceProperties().getProjectStartDateCode().equalsIgnoreCase(field.getCode()) && field.getValue() != null && !field.getValue().isBlank()).map(ReferenceFieldModel::getValue).findFirst().ifPresent(rda::setStart);
	        project.getDefinition().getFields().stream().filter(field -> this.configuration.getRdaFileTransformerServiceProperties().getProjectEndDateCode().equalsIgnoreCase(field.getCode()) && field.getValue() != null && !field.getValue().isBlank()).map(ReferenceFieldModel::getValue).findFirst().ifPresent(rda::setEnd);
        }
        if (!funders.isEmpty()) rda.setFunding(List.of(this.buildRRdaFunding(funders.getFirst(), grants.isEmpty() ? null : grants.getFirst())));

        return rda;
    }

    public List<Project> buildRdaProjects(List<ReferenceModel> projects, List<ReferenceModel> grants,  List<ReferenceModel> funder) {
        List<Project> rdaModels = new ArrayList<>();
        if (projects == null) return rdaModels;
        for (ReferenceModel userDMP : projects){
            rdaModels.add(this.buildRdaProject(userDMP, grants, funder));
        }
        return rdaModels;
    }

    public Funding buildRRdaFunding(ReferenceModel funder, ReferenceModel grant) {
        if (grant == null && funder == null) return null;
        Funding rda = new Funding();
        if (funder != null && funder.getReference() != null) {
            FunderId funderId = new FunderId();
            funderId.setIdentifier(funder.getReference());
            funderId.setType(FunderId.Type.FUNDREF);
            rda.setFunderId(funderId);
        } else if (funder != null){
            FunderId funderId = new FunderId();
            funderId.setIdentifier(funder.getId().toString());
            funderId.setType(FunderId.Type.OTHER);
            rda.setFunderId(funderId);
        }
        if (grant != null && grant.getReference() != null) {
            GrantId grantId = new GrantId();
            grantId.setIdentifier(grant.getReference());
            grantId.setType(GrantId.Type.OTHER);
            rda.setGrantId(grantId);
        } else if (grant != null) {
            GrantId grantId = new GrantId();
            grantId.setIdentifier(grant.getId().toString());
            grantId.setType(GrantId.Type.OTHER);
            rda.setGrantId(grantId);
        }
        return rda;
    }


    public Dmp.EthicalIssuesExist buildRdaEthicalIssuesExist(DmpModel model) {
        if (model == null) throw new MyApplicationException("Dmp is missing");
        List<org.opencdmp.commonmodels.models.dmpblueprint.FieldModel> fields = this.getFieldsOfSemantic(model, SEMANTIC_ETHICAL_ISSUES_EXISTS);
        for(org.opencdmp.commonmodels.models.dmpblueprint.FieldModel field : fields) {
            DmpBlueprintValueModel valueModel = field != null ? this.getDmpBlueprintValue(model, field.getId()) : null;
            if (valueModel != null && valueModel.getValue() != null && !valueModel.getValue().isBlank()) return Dmp.EthicalIssuesExist.fromValue(valueModel.getValue());
        }
        return Dmp.EthicalIssuesExist.UNKNOWN;
    }

    public URI buildRdaEthicalIssueReport(DmpModel model) {
        if (model == null) throw new MyApplicationException("Dmp is missing");
        List<org.opencdmp.commonmodels.models.dmpblueprint.FieldModel> fields = this.getFieldsOfSemantic(model, SEMANTIC_ETHICAL_ISSUES_REPORT);
        for(org.opencdmp.commonmodels.models.dmpblueprint.FieldModel field : fields) {
            DmpBlueprintValueModel valueModel = field != null ? this.getDmpBlueprintValue(model, field.getId()) : null;
            if (valueModel != null && valueModel.getValue() != null && !valueModel.getValue().isBlank()) return URI.create(valueModel.getValue());
        }
        return null;
    }

    public String buildRdaEthicalIssuesDescription(DmpModel model) {
        if (model == null) throw new MyApplicationException("Dmp is missing");
        List<org.opencdmp.commonmodels.models.dmpblueprint.FieldModel> fields = this.getFieldsOfSemantic(model, SEMANTIC_ETHICAL_ISSUES_DESCRIPTION);
        for(org.opencdmp.commonmodels.models.dmpblueprint.FieldModel field : fields) {
            DmpBlueprintValueModel valueModel = field != null ? this.getDmpBlueprintValue(model, field.getId()) : null;
            if (valueModel != null && valueModel.getValue() != null && !valueModel.getValue().isBlank()) return valueModel.getValue();
        }
        return null;
    }

    public Contact buildRdaContact(DmpModel model) {
        if (model == null) throw new MyApplicationException("Dmp is missing");
        UserModel creator = model.getCreator();
        if (creator == null) {
            creator = model.getUsers().stream().filter(x -> x.getRole().equals(DmpUserRole.Owner) && x.getUser() != null && x.getUser().getContacts() != null && x.getUser().getContacts()
                    .stream().anyMatch(y -> y.getType().equals(ContactInfoType.Email) && y.getValue() != null && !y.getValue().isBlank())).map(DmpUserModel::getUser).findFirst().orElse(null);
        }

        if (creator == null) throw new MyApplicationException("Creator Name is missing");
        if (creator.getName() == null) throw new MyApplicationException("Contact Name is missing");
        UserContactInfoModel emailContact = creator.getContacts() != null ? creator.getContacts().stream().filter(userContactInfo -> userContactInfo.getType().equals(ContactInfoType.Email)).findFirst().orElse(null) : null;
        if (emailContact == null)  throw new MyApplicationException("Contact Email is missing");

        Contact rdaModel = new Contact();
        rdaModel.setName(creator.getName());

        rdaModel.setMbox(emailContact.getValue());
        ContactId rdaContactId = new ContactId();
        rdaContactId.setIdentifier(emailContact.getId().toString());
        rdaContactId.setType(ContactId.Type.OTHER);
        rdaModel.setContactId(rdaContactId);
        
        return rdaModel;
    }

    public List<Contributor> buildRdaContributorsByUsers(List<DmpUserModel> models) {
        List<Contributor> rdaModels = new ArrayList<>();
        if (models == null) return rdaModels;
        for (DmpUserModel userDMP : models){
            rdaModels.add(this.buildRdaContributorsByUser(userDMP));
        }
        return rdaModels;
    }

    public Contributor buildRdaContributorsByUser(DmpUserModel model) {
        if (model == null) throw new MyApplicationException("User is missing");
        if (model.getUser() == null) throw new MyApplicationException("User is missing");
        if (model.getUser().getName() == null) throw new MyApplicationException("Contributor Name is missing");

        Contributor rdaModel = new Contributor();
        rdaModel.setContributorId(buildRdaContributorId(model.getUser().getId().toString()));

        rdaModel.setName(model.getUser().getName());
        UserContactInfoModel emailContact = model.getUser().getContacts() == null ? null : model.getUser().getContacts().stream().filter(userContactInfo -> userContactInfo.getType().equals(ContactInfoType.Email)).findFirst().orElse(null);
        if (emailContact != null) {
            rdaModel.setMbox(emailContact.getValue());
        }
        rdaModel.setRole(new HashSet<>(List.of(model.getRole().name())));
        return rdaModel;
    }

    public List<Contributor> buildRdaContributors(List<ReferenceModel> models) {
        List<Contributor> rdaModels = new ArrayList<>();
        if (models == null) return rdaModels;
        for (ReferenceModel researcher : models){
            Contributor item = this.buildRdaContributor(researcher);
            if (item != null) rdaModels.add(item);
        }
        return rdaModels;
    }

    public Contributor buildRdaContributor(ReferenceModel model) {
        if (model == null) throw new MyApplicationException("Contributor is missing");

        Contributor rdaModel = new Contributor();
        rdaModel.setContributorId(buildRdaContributorId(model.getReference()));
        rdaModel.setName(model.getLabel());
        if (model.getDefinition() != null && model.getDefinition().getFields() != null) {
	        model.getDefinition().getFields().stream().filter(field -> this.configuration.getRdaFileTransformerServiceProperties().getResearcherMalCode().equalsIgnoreCase(field.getCode())).findFirst().ifPresent(emailField -> rdaModel.setMbox(emailField.getValue()));
        }
        return rdaModel;
    }

    public ContributorId buildRdaContributorId(String id) {
        if (id == null || id.isBlank()) throw  new MyApplicationException("ContributorId is missing"); 
        ContributorId rda = new ContributorId();
        String[] idParts = id.split(":");
        String prefix = idParts.length > 1 ? idParts[0] : id;
        if (prefix.equals(PREFIX_ORCID)) {
            String finalId = id.replace(prefix + ":", "");
            rda.setIdentifier(ORCID_URL + finalId);
            rda.setType(ContributorId.Type.ORCID);
        } else {
            rda.setIdentifier(id);
            rda.setType(ContributorId.Type.OTHER);
        }
        return rda;
    }

    private DmpId buildRdaDmpId(DmpModel dmp){
        if (dmp == null) throw new MyApplicationException("DMP is missing");
        EntityDoiModel model = dmp.getEntityDois() == null ? null : dmp.getEntityDois().stream().filter(x -> x.getRepositoryId() != null && x.getDoi() != null && !x.getDoi().isBlank() && x.getRepositoryId().equalsIgnoreCase(REPOSITORY_ZENODO)).findFirst().orElse(null);
        DmpId rdaModel = new DmpId();
        if (model == null){
            rdaModel.setIdentifier(dmp.getId().toString());
            rdaModel.setType(DmpId.Type.OTHER);
        } else {
            rdaModel.setIdentifier(model.getDoi());
            rdaModel.setType(DmpId.Type.DOI);
        }
        
        return rdaModel;
    }

    private Language buildRdaLanguage(String language){
        if (language == null) throw new MyApplicationException("Language is missing");
        String rdaLanguage = this.configuration.getRdaFileTransformerServiceProperties().getLanguageMap().getOrDefault(language, null);

        return rdaLanguage != null ? Language.fromValue(rdaLanguage) : null;
    }

    private List<ReferenceModel> getReferenceModelOfTypeCode(DmpModel dmp, String code){
        List<ReferenceModel> response = new ArrayList<>();
        if (dmp.getReferences() == null) return response;
        for (DmpReferenceModel dmpReferenceModel : dmp.getReferences()){
            if (dmpReferenceModel.getReference() != null && dmpReferenceModel.getReference().getType() != null && dmpReferenceModel.getReference().getType().getCode() != null  && dmpReferenceModel.getReference().getType().getCode().equals(code)){
                response.add(dmpReferenceModel.getReference());
            }
        }

        return response;
    }

    private List<org.opencdmp.commonmodels.models.dmpblueprint.FieldModel> getFieldsOfSemantic(DmpModel dmp, String semanticKey){
        if (dmp == null || dmp.getDmpBlueprint() == null || dmp.getDmpBlueprint().getDefinition() == null || dmp.getDmpBlueprint().getDefinition().getSections() == null) return null;
        List<org.opencdmp.commonmodels.models.dmpblueprint.FieldModel> fieldModels = new ArrayList<>();
        for (SectionModel sectionModel : dmp.getDmpBlueprint().getDefinition().getSections()){
            if (sectionModel.getFields() != null){
	            sectionModel.getFields().stream().filter(x -> x.getSemantics() != null && x.getSemantics().contains(semanticKey)).findFirst().ifPresent(fieldModels::add);
            }
        }
        return fieldModels;
    }

    private DmpBlueprintValueModel getDmpBlueprintValue(DmpModel dmp, UUID id){
        if (dmp == null || dmp.getProperties() == null || dmp.getProperties().getDmpBlueprintValues() == null) return null;
        return dmp.getProperties().getDmpBlueprintValues().stream().filter(x-> x.getFieldId().equals(id)).findFirst().orElse(null);
    }
    
    //endregion
    
    //region dataset

    private Dataset buildRdaDataset(DescriptionModel model, Dmp dmpRda){
        if (model == null) throw new MyApplicationException("Model is missing");
        if (model.getLabel() == null)  throw new MyApplicationException("Dataset Label is missing");

        Dataset rda = new Dataset();
        rda.setDatasetId(this.buildRdaDatasetId(model));
        rda.setTitle(model.getLabel());
        rda.setDescription(model.getDescription());
        rda.setLanguage(this.buildRdaDatasetLanguage(model));
        rda.setType(this.buildRdaDatasetType(model));
        rda.setIssued(this.buildRdaDatasetIssued(model));
        rda.setKeyword(this.buildRdaDatasetKeywords(model));
        if (rda.getKeyword().isEmpty() && model.getTags() != null){
            rda.getKeyword().addAll(model.getTags());
        }
        rda.setPersonalData(this.buildRdaDatasetPersonalData(model));
        rda.setSensitiveData(this.buildRdaDatasetSensitiveData(model));
        rda.setPreservationStatement(this.buildRdaDatasetPreservationStatement(model));
        rda.setTechnicalResource(this.buildRdaDatasetTechnicalResources(model));
        rda.setSecurityAndPrivacy(this.buildRdaDatasetSecurityAndPrivacy(model));
        rda.setMetadata(this.buildRdaDatasetMetadata(model));
        rda.setDataQualityAssurance(this.buildDataQualityAssurances(model));
        rda.setDistribution(this.buildRdaDistribution(model));

        dmpRda.getCost().addAll(this.buildRdaCosts(model));
        dmpRda.getContributor().addAll(this.buildContributor(model));
        this.mergeEthicalIssues(model, dmpRda);
            
        return rda;
    }

    private List<Contributor> buildContributor(DescriptionModel model){
        List<Contributor> values = new ArrayList<>();
        for (FieldModel field : this.templateFieldSearcherService.searchFieldsBySemantics(model.getDescriptionTemplate(), SEMANTIC_DATASET_CONTRIBUTOR)) {
            List<org.opencdmp.commonmodels.models.description.FieldModel> fieldValues = this.findValueField(field, model.getProperties());
            for (org.opencdmp.commonmodels.models.description.FieldModel fieldValue : fieldValues) {
                if (fieldValue.getReferences() != null  && !fieldValue.getReferences().isEmpty()) {
                    for (ReferenceModel referenceModel : fieldValue.getReferences()){
                        values.add(this.buildRdaContributor(referenceModel));
                    }
                }
            }
        }
        return values;
    }
    
    
    private void mergeEthicalIssues(DescriptionModel model, Dmp dmpRda){
        for (FieldModel field : this.templateFieldSearcherService.searchFieldsBySemantics(model.getDescriptionTemplate(), SEMANTIC_ETHICAL_ISSUES_EXISTS)) {
            List<org.opencdmp.commonmodels.models.description.FieldModel> fieldValues = this.findValueField(field, model.getProperties());
            for (org.opencdmp.commonmodels.models.description.FieldModel fieldValue : fieldValues) {
                if (fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
                    try {
                        Dmp.EthicalIssuesExist value = Dmp.EthicalIssuesExist.fromValue(fieldValue.getTextValue());
                        if (!Dmp.EthicalIssuesExist.UNKNOWN.equals(value)) dmpRda.setEthicalIssuesExist(value);
                    } catch (Exception e) {
                        logger.error(e.getMessage(), e);
                    }
                }
            }
        }
        for (FieldSetModel fieldSet : this.templateFieldSearcherService.searchFieldSetsBySemantics(model.getDescriptionTemplate(), List.of(SEMANTIC_ETHICAL_ISSUES_EXISTS, SEMANTIC_ETHICAL_ISSUES_DESCRIPTION, SEMANTIC_ETHICAL_ISSUES_REPORT))) {
            List<org.opencdmp.commonmodels.models.description.PropertyDefinitionFieldSetItemModel> propertyDefinitionFieldSetItemModels = this.findFieldSetValue(fieldSet, model.getProperties());
            for (org.opencdmp.commonmodels.models.description.PropertyDefinitionFieldSetItemModel propertyDefinitionFieldSetItemModel : propertyDefinitionFieldSetItemModels) {

                org.opencdmp.commonmodels.models.description.FieldModel fieldValue = this.findValueFieldBySemantic(fieldSet, propertyDefinitionFieldSetItemModel, SEMANTIC_ETHICAL_ISSUES_EXISTS);
                if (fieldValue != null && fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
                    try {
                        Dmp.EthicalIssuesExist value = Dmp.EthicalIssuesExist.fromValue(fieldValue.getTextValue());
                        if (!Dmp.EthicalIssuesExist.UNKNOWN.equals(value)) dmpRda.setEthicalIssuesExist(value);
                    } catch (Exception e) { logger.error(e.getMessage(), e); }
                }
                if (fieldValue != null && fieldValue.getTextListValue() != null && !fieldValue.getTextListValue().isEmpty()) {
                    for (String val : fieldValue.getTextListValue()) {
                        try {
                            Dmp.EthicalIssuesExist value = Dmp.EthicalIssuesExist.fromValue(val);
                            if (!Dmp.EthicalIssuesExist.UNKNOWN.equals(value)) dmpRda.setEthicalIssuesExist(value);
                        } catch (Exception e) {
                            logger.error(e.getMessage(), e);
                        }
                    }
                }

                if (dmpRda.getEthicalIssuesDescription() == null || dmpRda.getEthicalIssuesDescription().isBlank()) {
                    fieldValue = this.findValueFieldBySemantic(fieldSet, propertyDefinitionFieldSetItemModel, SEMANTIC_ETHICAL_ISSUES_DESCRIPTION);
                    if (fieldValue != null && fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
                        dmpRda.setEthicalIssuesDescription(fieldValue.getTextValue());
                    }
                }

                if (dmpRda.getEthicalIssuesReport() == null) {
                    fieldValue = this.findValueFieldBySemantic(fieldSet, propertyDefinitionFieldSetItemModel, SEMANTIC_ETHICAL_ISSUES_REPORT);
                    if (fieldValue != null && fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
                        try {
                            dmpRda.setEthicalIssuesReport(URI.create(fieldValue.getTextValue()));
                        } catch (Exception ex) {
                            logger.error(ex.getMessage(), ex);
                        }
                    }
                }
            }
        }
    }

    private List<Cost> buildRdaCosts(DescriptionModel model){
        List<Cost> values = new ArrayList<>();
        for (FieldSetModel fieldSet : this.templateFieldSearcherService.searchFieldSetsBySemantics(model.getDescriptionTemplate(), List.of(SEMANTIC_COST_CURRENCY_CODE, SEMANTIC_COST_VALUE, SEMANTIC_COST_DESCRIPTION, SEMANTIC_COST_TITLE))) {
            List<org.opencdmp.commonmodels.models.description.PropertyDefinitionFieldSetItemModel> propertyDefinitionFieldSetItemModels = this.findFieldSetValue(fieldSet, model.getProperties());
            for (org.opencdmp.commonmodels.models.description.PropertyDefinitionFieldSetItemModel propertyDefinitionFieldSetItemModel : propertyDefinitionFieldSetItemModels) {
                boolean valueFound = false;
                Cost item = new Cost();

                org.opencdmp.commonmodels.models.description.FieldModel fieldValue = this.findValueFieldBySemantic(fieldSet, propertyDefinitionFieldSetItemModel, SEMANTIC_COST_CURRENCY_CODE);
                if (fieldValue != null && fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
                    try {
                        item.setCurrencyCode(Cost.CurrencyCode.fromValue(fieldValue.getTextValue()));
                        valueFound = true;
                    } catch (Exception e) {
                        logger.error(e.getMessage(), e);
                    }
                }

                fieldValue = this.findValueFieldBySemantic(fieldSet, propertyDefinitionFieldSetItemModel, SEMANTIC_COST_DESCRIPTION);
                if (fieldValue != null && fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
                    item.setDescription(fieldValue.getTextValue());
                    valueFound = true;
                }

                fieldValue = this.findValueFieldBySemantic(fieldSet, propertyDefinitionFieldSetItemModel, SEMANTIC_COST_TITLE);
                if (fieldValue != null && fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
                    item.setTitle(fieldValue.getTextValue());
                    valueFound = true;
                }

                fieldValue = this.findValueFieldBySemantic(fieldSet, propertyDefinitionFieldSetItemModel, SEMANTIC_COST_VALUE);
                if (fieldValue != null && fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
                    try {
                        item.setValue(Double.valueOf(fieldValue.getTextValue()));
                        valueFound = true;
                    } catch (Exception ex) {
                        logger.error(ex.getMessage(), ex);
                    }
                }

                if (valueFound) values.add(item);
            }
        }
        return values;
    }

    private List<Distribution> buildRdaDistribution(DescriptionModel model){
        List<Distribution> values = new ArrayList<>();
        for (FieldSetModel fieldSet : this.templateFieldSearcherService.searchFieldSetsBySemantics(model.getDescriptionTemplate(), List.of(
                SEMANTIC_DATASET_DISTRIBUTION_LICENCE, SEMANTIC_DATASET_DISTRIBUTION_TITLE, SEMANTIC_DATASET_DISTRIBUTION_FORMAT,
                SEMANTIC_DATASET_DISTRIBUTION_DOWNLOAD_URL, SEMANTIC_DATASET_DISTRIBUTION_DESCRIPTION, SEMANTIC_DATASET_DISTRIBUTION_DATA_ACCESS,
                SEMANTIC_DATASET_DISTRIBUTION_BYTE_SIZE,  SEMANTIC_DATASET_DISTRIBUTION_AVAILABLE_UTIL, SEMANTIC_DATASET_DISTRIBUTION_ACCESS_URL))) {
            List<org.opencdmp.commonmodels.models.description.PropertyDefinitionFieldSetItemModel> propertyDefinitionFieldSetItemModels = this.findFieldSetValue(fieldSet, model.getProperties());
            for (org.opencdmp.commonmodels.models.description.PropertyDefinitionFieldSetItemModel propertyDefinitionFieldSetItemModel : propertyDefinitionFieldSetItemModels) {
                boolean valueFound = false;
                Distribution item = new Distribution();

                org.opencdmp.commonmodels.models.description.FieldModel fieldValue = this.findValueFieldBySemantic(fieldSet, propertyDefinitionFieldSetItemModel, SEMANTIC_DATASET_DISTRIBUTION_TITLE);
                if (fieldValue != null && fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
                    item.setTitle(fieldValue.getTextValue());
                    valueFound = true;
                }

                fieldValue = this.findValueFieldBySemantic(fieldSet, propertyDefinitionFieldSetItemModel, SEMANTIC_DATASET_DISTRIBUTION_FORMAT);
                item.setFormat(new ArrayList<>());
                if (fieldValue != null && fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) { 
                    item.getFormat().add(fieldValue.getTextValue());
                    valueFound = true;
                }
                if (fieldValue != null && fieldValue.getTextListValue() != null && !fieldValue.getTextListValue().isEmpty()) {
                    item.getFormat().addAll(fieldValue.getTextListValue());
                    valueFound = true;
                }

                fieldValue = this.findValueFieldBySemantic(fieldSet, propertyDefinitionFieldSetItemModel, SEMANTIC_DATASET_DISTRIBUTION_DOWNLOAD_URL);
                if (fieldValue != null && fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
                    try {
                        item.setDownloadUrl(URI.create(fieldValue.getTextValue()));
                        valueFound = true;
                    } catch (Exception ex) { logger.error(ex.getMessage(), ex); }
                }

                fieldValue = this.findValueFieldBySemantic(fieldSet, propertyDefinitionFieldSetItemModel, SEMANTIC_DATASET_DISTRIBUTION_DESCRIPTION);
                if (fieldValue != null && fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
                    item.setDescription(fieldValue.getTextValue());
                    valueFound = true;
                }

                fieldValue = this.findValueFieldBySemantic(fieldSet, propertyDefinitionFieldSetItemModel, SEMANTIC_DATASET_DISTRIBUTION_DATA_ACCESS);
                if (fieldValue != null && fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
                    try {
                        item.setDataAccess(Distribution.DataAccess.fromValue(fieldValue.getTextValue()));
                        valueFound = true;
                    } catch (Exception ex) { logger.error(ex.getMessage(), ex); }
                }
                if (fieldValue != null && fieldValue.getTextListValue() != null && !fieldValue.getTextListValue().isEmpty()) {
                    for (String val : fieldValue.getTextListValue()) {
                        try {
                            item.setDataAccess(Distribution.DataAccess.fromValue(val));
                            valueFound = true;
                        } catch (Exception ex) {
                            logger.error(ex.getMessage(), ex);
                        }
                    }
                    
                }

                fieldValue = this.findValueFieldBySemantic(fieldSet, propertyDefinitionFieldSetItemModel, SEMANTIC_DATASET_DISTRIBUTION_BYTE_SIZE);
                if (fieldValue != null && fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
                    try {
                        item.setByteSize(Integer.valueOf(fieldValue.getTextValue()));
                        valueFound = true;
                    } catch (Exception ex) { logger.error(ex.getMessage(), ex); }
                }

                fieldValue = this.findValueFieldBySemantic(fieldSet, propertyDefinitionFieldSetItemModel, SEMANTIC_DATASET_DISTRIBUTION_AVAILABLE_UTIL);
                if (fieldValue != null && fieldValue.getDateValue() != null) {
//                    item.setAvailableUntil(DateTimeFormatter.ISO_DATE_TIME.withZone(ZoneId.systemDefault()).format(fieldValue.getDateValue()));
                    item.setAvailableUntil(fieldValue.getDateValue().toString());
                    valueFound = true;
                } else if (fieldValue != null && fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
                    item.setAvailableUntil(fieldValue.getTextValue());
                    valueFound = true;
                }

                fieldValue = this.findValueFieldBySemantic(fieldSet, propertyDefinitionFieldSetItemModel, SEMANTIC_DATASET_DISTRIBUTION_ACCESS_URL);
                if (fieldValue != null && fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
                    item.setAccessUrl(fieldValue.getTextValue());
                    valueFound = true;
                }

                fieldValue = this.findValueFieldBySemantic(fieldSet, propertyDefinitionFieldSetItemModel, SEMANTIC_DATASET_DISTRIBUTION_ACCESS_URL);
                if (fieldValue != null && fieldValue.getReferences() != null && !fieldValue.getReferences().isEmpty()) {
                    item.setLicense(new ArrayList<>());
                    for (ReferenceModel referenceModel : fieldValue.getReferences()) {
                        if (referenceModel.getDefinition() != null && referenceModel.getDefinition().getFields() != null) {
                            try {
                                License license = new License();
                                String refValue = referenceModel.getReference();
                                if (refValue != null && !refValue.isBlank()) license.setLicenseRef(URI.create(refValue));
                                String startDateValue = referenceModel.getDefinition().getFields().stream().filter(f -> this.configuration.getRdaFileTransformerServiceProperties().getLicenseStartDateCode().equalsIgnoreCase(f.getCode()) && f.getValue() != null && !f.getValue().isBlank()).map(ReferenceFieldModel::getValue).findFirst().orElse(null);
                                if (refValue != null && !refValue.isBlank()) license.setStartDate(startDateValue);
                                else license.setStartDate(Instant.now().toString());

                                item.getLicense().add(license);
                            } catch (Exception ex) {
                                logger.error(ex.getMessage(), ex);
                            }
                        }
                    }
                    if (!item.getLicense().isEmpty()) valueFound = true;
                }

                if (valueFound) {
                    item.setHost(this.buildRdaDistributionHost(propertyDefinitionFieldSetItemModel, fieldSet));
                    values.add(item);
                }
            }
        }
        return values.isEmpty() ? null : values;
    }

    private Host buildRdaDistributionHost(org.opencdmp.commonmodels.models.description.PropertyDefinitionFieldSetItemModel propertyDefinitionFieldSetItemModel, FieldSetModel fieldSet) {
        boolean valueFound = false;
        Host item = new Host();
        org.opencdmp.commonmodels.models.description.FieldModel fieldValue = this.findValueFieldBySemantic(fieldSet, propertyDefinitionFieldSetItemModel, SEMANTIC_DATASET_DISTRIBUTION_HOST_TITLE);
        if (fieldValue != null && fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
            item.setTitle(fieldValue.getTextValue());
            valueFound = true;
        }

        fieldValue = this.findValueFieldBySemantic(fieldSet, propertyDefinitionFieldSetItemModel, SEMANTIC_DATASET_DISTRIBUTION_HOST_URL);
        if (fieldValue != null && fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
            try {
                item.setUrl(URI.create(fieldValue.getTextValue()));
                valueFound = true;
            } catch (Exception ex) { logger.error(ex.getMessage(), ex); }
        }

        fieldValue = this.findValueFieldBySemantic(fieldSet, propertyDefinitionFieldSetItemModel, SEMANTIC_DATASET_DISTRIBUTION_HOST_DESCRIPTION);
        if (fieldValue != null && fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
            item.setDescription(fieldValue.getTextValue());
            valueFound = true;
        }

        fieldValue = this.findValueFieldBySemantic(fieldSet, propertyDefinitionFieldSetItemModel, SEMANTIC_DATASET_DISTRIBUTION_HOST_STORAGE_TYPE);
        if (fieldValue != null && fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
            item.setStorageType(fieldValue.getTextValue());
            valueFound = true;
        }

        fieldValue = this.findValueFieldBySemantic(fieldSet, propertyDefinitionFieldSetItemModel, SEMANTIC_DATASET_DISTRIBUTION_HOST_BACKUP_TYPE);
        if (fieldValue != null && fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
            item.setBackupType(fieldValue.getTextValue());
            valueFound = true;
        }

        fieldValue = this.findValueFieldBySemantic(fieldSet, propertyDefinitionFieldSetItemModel, SEMANTIC_DATASET_DISTRIBUTION_HOST_AVAILABILITY);
        if (fieldValue != null && fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
            item.setAvailability(fieldValue.getTextValue());
            valueFound = true;
        }

        fieldValue = this.findValueFieldBySemantic(fieldSet, propertyDefinitionFieldSetItemModel, SEMANTIC_DATASET_DISTRIBUTION_HOST_BACKUP_FREQUENCY);
        if (fieldValue != null && fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
            item.setBackupFrequency(fieldValue.getTextValue());
            valueFound = true;
        }

        fieldValue = this.findValueFieldBySemantic(fieldSet, propertyDefinitionFieldSetItemModel, SEMANTIC_DATASET_DISTRIBUTION_HOST_GEO_LOCATION);
        if (fieldValue != null && fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
            try {
                item.setGeoLocation(Host.GeoLocation.fromValue(fieldValue.getTextValue()));
                valueFound = true;
            } catch (Exception ex) { logger.error(ex.getMessage(), ex); }
        }
        if (fieldValue != null && fieldValue.getTextListValue() != null && !fieldValue.getTextListValue().isEmpty()) {
            for (String val : fieldValue.getTextListValue()) {
                try {
                    item.setGeoLocation(Host.GeoLocation.fromValue(val));
                    valueFound = true;
                } catch (Exception ex) {
                    logger.error(ex.getMessage(), ex);
                }
            }
        }

        fieldValue = this.findValueFieldBySemantic(fieldSet, propertyDefinitionFieldSetItemModel, SEMANTIC_DATASET_DISTRIBUTION_HOST_SUPPORT_VERSIONING);
        if (fieldValue != null && fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
            try {
                item.setSupportVersioning(Host.SupportVersioning.fromValue(fieldValue.getTextValue()));
                valueFound = true;
            } catch (Exception ex) {
                logger.error(ex.getMessage(), ex);
            }
        }
        if (fieldValue != null && fieldValue.getTextListValue() != null && !fieldValue.getTextListValue().isEmpty()) {
            for (String val : fieldValue.getTextListValue()) {
                try {
                    item.setSupportVersioning(Host.SupportVersioning.fromValue(val));
                    valueFound = true;
                } catch (Exception ex) {
                    logger.error(ex.getMessage(), ex);
                }
            }
        }

        fieldValue = this.findValueFieldBySemantic(fieldSet, propertyDefinitionFieldSetItemModel, SEMANTIC_DATASET_DISTRIBUTION_HOST_CERTIFIED_WITH);
        if (fieldValue != null && fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
            try {
                item.setCertifiedWith(Host.CertifiedWith.fromValue(fieldValue.getTextValue()));
                valueFound = true;
            } catch (Exception ex) {
                logger.error(ex.getMessage(), ex);
            }
        }
        if (fieldValue != null && fieldValue.getTextListValue() != null && !fieldValue.getTextListValue().isEmpty()) {
            for (String val : fieldValue.getTextListValue()) {
                try {
                    item.setCertifiedWith(Host.CertifiedWith.fromValue(val));
                    valueFound = true;
                } catch (Exception ex) {
                    logger.error(ex.getMessage(), ex);
                }
            }
        }
        
        fieldValue = this.findValueFieldBySemantic(fieldSet, propertyDefinitionFieldSetItemModel, SEMANTIC_DATASET_DISTRIBUTION_HOST_PID_SYSTEM);
        item.setPidSystem(new ArrayList<>());
        if (fieldValue != null && fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
            try {
                item.getPidSystem().add(PidSystem.fromValue(fieldValue.getTextValue()));
                valueFound = true;
            } catch (Exception ex) {
                logger.error(ex.getMessage(), ex);
            }
        }
        if (fieldValue != null && fieldValue.getTextListValue() != null && !fieldValue.getTextListValue().isEmpty()) {
            for (String itm : fieldValue.getTextListValue()) {
                try {

                    item.getPidSystem().add(PidSystem.fromValue(itm));
                    valueFound = true;
                } catch (Exception ex) {
                    logger.error(ex.getMessage(), ex);
                }
            }
        }

        if (valueFound) return item;
        return null;
    }

    private List<String> buildDataQualityAssurances(DescriptionModel model){
        List<String> values = new ArrayList<>();
        for (FieldModel field : this.templateFieldSearcherService.searchFieldsBySemantics(model.getDescriptionTemplate(), SEMANTIC_DATASET_DATA_QUALITY_ASSURANCE)) {
            List<org.opencdmp.commonmodels.models.description.FieldModel> fieldValues = this.findValueField(field, model.getProperties());
            for (org.opencdmp.commonmodels.models.description.FieldModel fieldValue : fieldValues) {
                if (fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
                    values.add(fieldValue.getTextValue());
                }
                if (fieldValue.getTextListValue() != null && !fieldValue.getTextListValue().isEmpty()) values.addAll(fieldValue.getTextListValue());
            }
        }
        return values;
    }

    private DatasetId buildRdaDatasetId(DescriptionModel model){
        DatasetId datasetId = new DatasetId();
        datasetId.setType(DatasetId.Type.OTHER);
        
        for (FieldModel field : this.templateFieldSearcherService.searchFieldsBySemantics(model.getDescriptionTemplate(), SEMANTIC_DATASET_DATASET_ID)) {
            List<org.opencdmp.commonmodels.models.description.FieldModel> fieldValues = this.findValueField(field, model.getProperties());
            for (org.opencdmp.commonmodels.models.description.FieldModel fieldValue : fieldValues) {
                if (fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
                    datasetId.setIdentifier(fieldValue.getTextValue() );
                    return datasetId;
                }
            }
        }
        
        for (FieldSetModel fieldSet : this.templateFieldSearcherService.searchFieldSetsBySemantics(model.getDescriptionTemplate(), List.of(SEMANTIC_DATASET_DATASET_ID_ID, SEMANTIC_DATASET_DATASET_ID_TYPE))) {
            List<org.opencdmp.commonmodels.models.description.PropertyDefinitionFieldSetItemModel> propertyDefinitionFieldSetItemModels = this.findFieldSetValue(fieldSet, model.getProperties());
            for (org.opencdmp.commonmodels.models.description.PropertyDefinitionFieldSetItemModel propertyDefinitionFieldSetItemModel : propertyDefinitionFieldSetItemModels) {
                boolean valueFound = false;

                org.opencdmp.commonmodels.models.description.FieldModel fieldValue = this.findValueFieldBySemantic(fieldSet, propertyDefinitionFieldSetItemModel, SEMANTIC_DATASET_DATASET_ID_ID);
                if (fieldValue != null && fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
                    datasetId.setIdentifier(fieldValue.getTextValue());
                    valueFound = true;
                }

                fieldValue = this.findValueFieldBySemantic(fieldSet, propertyDefinitionFieldSetItemModel, SEMANTIC_DATASET_DATASET_ID_TYPE);
                if (fieldValue != null && fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
                    try {
                        datasetId.setType(DatasetId.Type.fromValue(fieldValue.getTextValue()));
                        valueFound = true;
                    } catch (Exception ex){
                        logger.error(ex.getMessage(), ex);
                    }
                }
                if (fieldValue != null && fieldValue.getTextListValue() != null && !fieldValue.getTextListValue().isEmpty()) {
                    for (String val : fieldValue.getTextListValue()) {
                        try {
                            datasetId.setType(DatasetId.Type.fromValue(val));
                            valueFound = true;
                        } catch (Exception ex) {
                            logger.error(ex.getMessage(), ex);
                        }
                    }
                }

                if (valueFound) return datasetId;
            }
        }
        return new DatasetId(model.getId().toString(), DatasetId.Type.OTHER);
    }

    private List<Metadatum> buildRdaDatasetMetadata(DescriptionModel model){
        List<Metadatum> values = new ArrayList<>();
        for (FieldSetModel fieldSet : this.templateFieldSearcherService.searchFieldSetsBySemantics(model.getDescriptionTemplate(), List.of(SEMANTIC_DATASET_METADATA_DESCRIPTION, SEMANTIC_DATASET_METADATA_LANGUAGE, SEMANTIC_DATASET_METADATA_STANDARD_ID, SEMANTIC_DATASET_METADATA_STANDARD_ID_IDENTIFIER, SEMANTIC_DATASET_METADATA_STANDARD_ID_TYPE))) {
            List<org.opencdmp.commonmodels.models.description.PropertyDefinitionFieldSetItemModel> propertyDefinitionFieldSetItemModels = this.findFieldSetValue(fieldSet, model.getProperties());
            for (org.opencdmp.commonmodels.models.description.PropertyDefinitionFieldSetItemModel propertyDefinitionFieldSetItemModel : propertyDefinitionFieldSetItemModels) {
                boolean valueFound = false;
                Metadatum item = new Metadatum();

                org.opencdmp.commonmodels.models.description.FieldModel fieldValue = this.findValueFieldBySemantic(fieldSet, propertyDefinitionFieldSetItemModel, SEMANTIC_DATASET_METADATA_LANGUAGE);
                if (fieldValue != null && fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
                    try {
                        item.setLanguage(Metadatum.Language.fromValue(fieldValue.getTextValue()));
                        valueFound = true;
                    } catch (Exception e) {
                        logger.error(e.getMessage(), e);
                    }
                }
                if (fieldValue != null && fieldValue.getTextListValue() != null && !fieldValue.getTextListValue().isEmpty()) {
                    for (String val : fieldValue.getTextListValue()) {
                        try {
                            item.setLanguage(Metadatum.Language.fromValue(val));
                            valueFound = true;
                        } catch (Exception e) {
                            logger.error(e.getMessage(), e);
                        }
                    }
                }
                

                fieldValue = this.findValueFieldBySemantic(fieldSet, propertyDefinitionFieldSetItemModel, SEMANTIC_DATASET_METADATA_DESCRIPTION);
                if (fieldValue != null && fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
                    item.setDescription(fieldValue.getTextValue());
                    valueFound = true;
                }
                MetadataStandardId standardId = this.buildRdaMetadataStandardId(propertyDefinitionFieldSetItemModel, fieldSet);
                if (standardId != null) {
                    item.setMetadataStandardId(standardId);
                    valueFound = true;

                }

                if (valueFound) values.add(item);
            }
        }
        return values.isEmpty() ? null : values;
    }

    private MetadataStandardId buildRdaMetadataStandardId(org.opencdmp.commonmodels.models.description.PropertyDefinitionFieldSetItemModel propertyDefinitionFieldSetItemModel, FieldSetModel fieldSet){
        boolean valueFound = false;
        MetadataStandardId standardId = new MetadataStandardId();
        standardId.setType(MetadataStandardId.Type.OTHER);

        org.opencdmp.commonmodels.models.description.FieldModel fieldValue = this.findValueFieldBySemantic(fieldSet, propertyDefinitionFieldSetItemModel, SEMANTIC_DATASET_METADATA_STANDARD_ID);
        if (fieldValue != null && fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
            standardId.setIdentifier(fieldValue.getTextValue());
            return standardId;
        }

        fieldValue = this.findValueFieldBySemantic(fieldSet, propertyDefinitionFieldSetItemModel, SEMANTIC_DATASET_METADATA_STANDARD_ID_IDENTIFIER);
        if (fieldValue != null && fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
            standardId.setIdentifier(fieldValue.getTextValue());
            valueFound = true;
        }

        fieldValue = this.findValueFieldBySemantic(fieldSet, propertyDefinitionFieldSetItemModel, SEMANTIC_DATASET_METADATA_STANDARD_ID_TYPE);
        if (fieldValue != null && fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
            try {
                standardId.setType(MetadataStandardId.Type.fromValue(fieldValue.getTextValue()));
            } catch (Exception ex){
                logger.error(ex.getMessage(), ex);
            }
        }
        if (fieldValue != null && fieldValue.getTextListValue() != null && !fieldValue.getTextListValue().isEmpty()) {
            for (String val : fieldValue.getTextListValue()) {
                try {
                    standardId.setType(MetadataStandardId.Type.fromValue(val));
                } catch (Exception ex) {
                    logger.error(ex.getMessage(), ex);
                }
            }
        }

        if (valueFound) return standardId;
        return null;
    }

    private List<SecurityAndPrivacy> buildRdaDatasetSecurityAndPrivacy(DescriptionModel model) {
        List<SecurityAndPrivacy> values = new ArrayList<>();
        for (FieldSetModel fieldSet : this.templateFieldSearcherService.searchFieldSetsBySemantics(model.getDescriptionTemplate(), List.of(SEMANTIC_DATASET_SECURITY_AND_PRIVACY_TITLE, SEMANTIC_DATASET_SECURITY_AND_PRIVACY_DESCRIPTION))) {
            List<org.opencdmp.commonmodels.models.description.PropertyDefinitionFieldSetItemModel> propertyDefinitionFieldSetItemModels = this.findFieldSetValue(fieldSet, model.getProperties());
            for (org.opencdmp.commonmodels.models.description.PropertyDefinitionFieldSetItemModel propertyDefinitionFieldSetItemModel : propertyDefinitionFieldSetItemModels) {
                boolean valueFound = false;
                SecurityAndPrivacy item = new SecurityAndPrivacy();

                org.opencdmp.commonmodels.models.description.FieldModel fieldValue = this.findValueFieldBySemantic(fieldSet, propertyDefinitionFieldSetItemModel, SEMANTIC_DATASET_SECURITY_AND_PRIVACY_TITLE);
                if (fieldValue != null && fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
                    item.setTitle(fieldValue.getTextValue());
                    valueFound = true;
                }

                fieldValue = this.findValueFieldBySemantic(fieldSet, propertyDefinitionFieldSetItemModel, SEMANTIC_DATASET_SECURITY_AND_PRIVACY_DESCRIPTION);
                if (fieldValue != null && fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
                    item.setDescription(fieldValue.getTextValue());
                    valueFound = true;
                }

                if (valueFound) values.add(item);
            }
        }
        return values.isEmpty() ? null : values;
    }
        
        

    private List<TechnicalResource> buildRdaDatasetTechnicalResources(DescriptionModel model) {
        List<TechnicalResource> values = new ArrayList<>();
        for (FieldSetModel fieldSet : this.templateFieldSearcherService.searchFieldSetsBySemantics(model.getDescriptionTemplate(), List.of(SEMANTIC_DATASET_TECHNICAL_RESOURCE_NAME, SEMANTIC_DATASET_TECHNICAL_RESOURCE_DESCRIPTION))) {
            List<org.opencdmp.commonmodels.models.description.PropertyDefinitionFieldSetItemModel> propertyDefinitionFieldSetItemModels = this.findFieldSetValue(fieldSet, model.getProperties());
            for (org.opencdmp.commonmodels.models.description.PropertyDefinitionFieldSetItemModel propertyDefinitionFieldSetItemModel : propertyDefinitionFieldSetItemModels) {
                boolean valueFound = false;
                TechnicalResource item = new TechnicalResource();

                org.opencdmp.commonmodels.models.description.FieldModel fieldValue = this.findValueFieldBySemantic(fieldSet, propertyDefinitionFieldSetItemModel, SEMANTIC_DATASET_TECHNICAL_RESOURCE_NAME);
                if (fieldValue != null && fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
                    item.setName(fieldValue.getTextValue());
                    valueFound = true;
                }

                fieldValue = this.findValueFieldBySemantic(fieldSet, propertyDefinitionFieldSetItemModel, SEMANTIC_DATASET_TECHNICAL_RESOURCE_DESCRIPTION);
                if (fieldValue != null && fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
                    item.setDescription(fieldValue.getTextValue());
                    valueFound = true;
                }

                if (valueFound) values.add(item);
            }
        }
        return values;

    }

    private String buildRdaDatasetPreservationStatement(DescriptionModel model){
        for (FieldModel field : this.templateFieldSearcherService.searchFieldsBySemantics(model.getDescriptionTemplate(), SEMANTIC_DATASET_PRESERVATION_STATEMENT)) {
            List<org.opencdmp.commonmodels.models.description.FieldModel> fieldValues = this.findValueField(field, model.getProperties());
            for (org.opencdmp.commonmodels.models.description.FieldModel fieldValue : fieldValues) {
                if (fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
                    return fieldValue.getTextValue();
                }
            }
        }
        return null;
    }

    private Dataset.SensitiveData buildRdaDatasetSensitiveData(DescriptionModel model){
        Dataset.SensitiveData value = null;
        for (FieldModel field : this.templateFieldSearcherService.searchFieldsBySemantics(model.getDescriptionTemplate(), SEMANTIC_DATASET_SENSITIVE_DATA)) {
            List<org.opencdmp.commonmodels.models.description.FieldModel> fieldValues = this.findValueField(field, model.getProperties());
            for (org.opencdmp.commonmodels.models.description.FieldModel fieldValue : fieldValues) {
                if (fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
                    try {
                        value = Dataset.SensitiveData.fromValue(fieldValue.getTextValue());
                        return value;
                    } catch (IllegalArgumentException | MyApplicationException e) {
                        logger.error(e.getMessage(), e);
                    }
                }
                if (fieldValue.getTextListValue() != null && !fieldValue.getTextListValue().isEmpty()) {
                    for (String val : fieldValue.getTextListValue()) {
                        try {
                            value = Dataset.SensitiveData.fromValue(val);
                            return value;
                        } catch (IllegalArgumentException | MyApplicationException e) {
                            logger.error(e.getMessage(), e);
                        }
                    }
                }
            }
        }
        return Dataset.SensitiveData.UNKNOWN;
    }
    
    private Dataset.PersonalData buildRdaDatasetPersonalData(DescriptionModel model){
        Dataset.PersonalData value = null;
        for (FieldModel field : this.templateFieldSearcherService.searchFieldsBySemantics(model.getDescriptionTemplate(), SEMANTIC_DATASET_PERSONAL_DATA)) {
            List<org.opencdmp.commonmodels.models.description.FieldModel> fieldValues = this.findValueField(field, model.getProperties());
            for (org.opencdmp.commonmodels.models.description.FieldModel fieldValue : fieldValues) {
                if (fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
                    try {
                        value = Dataset.PersonalData.fromValue(fieldValue.getTextValue());
                        return value;
                    } catch (IllegalArgumentException | MyApplicationException e) {
                        logger.error(e.getMessage(), e);
                    }
                }
                if (fieldValue.getTextListValue() != null && !fieldValue.getTextListValue().isEmpty()) {
                    for (String val : fieldValue.getTextListValue()) {
                        try {
                            value = Dataset.PersonalData.fromValue(val);
                            return value;
                        } catch (IllegalArgumentException | MyApplicationException e) {
                            logger.error(e.getMessage(), e);
                        }
                    }
                }
            }
        }
        return Dataset.PersonalData.UNKNOWN;
    }

    private List<String> buildRdaDatasetKeywords(DescriptionModel model){
        List<String> values = new ArrayList<>();
        for (FieldModel field : this.templateFieldSearcherService.searchFieldsBySemantics(model.getDescriptionTemplate(), SEMANTIC_DATASET_KEYWORD)) {
            List<org.opencdmp.commonmodels.models.description.FieldModel> fieldValues = this.findValueField(field, model.getProperties());
            for (org.opencdmp.commonmodels.models.description.FieldModel fieldValue : fieldValues) {
                if (fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
                    values.add(fieldValue.getTextValue());
                }
                if (fieldValue.getTextListValue() != null && !fieldValue.getTextListValue().isEmpty()) {
                    values.addAll(fieldValue.getTextListValue());
                }
            }
        }
        return values;
    }

    private String buildRdaDatasetIssued(DescriptionModel model){
        for (FieldModel field : this.templateFieldSearcherService.searchFieldsBySemantics(model.getDescriptionTemplate(), SEMANTIC_DATASET_ISSUED)) {
            List<org.opencdmp.commonmodels.models.description.FieldModel> fieldValues = this.findValueField(field, model.getProperties());
            for (org.opencdmp.commonmodels.models.description.FieldModel fieldValue : fieldValues) {
                if (fieldValue.getDateValue() != null){
//                    return DateTimeFormatter.ISO_DATE_TIME.withZone(ZoneId.systemDefault()).format(fieldValue.getDateValue());
                    return fieldValue.getDateValue().toString();
                } else if (fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
                    return fieldValue.getTextValue();
                }
            }
        }
        return null;
    }
    
    private Language buildRdaDatasetLanguage(DescriptionModel model){
        Language value = null;
        for (FieldModel field : this.templateFieldSearcherService.searchFieldsBySemantics(model.getDescriptionTemplate(), SEMANTIC_DATASET_LANGUAGE)) {
            List<org.opencdmp.commonmodels.models.description.FieldModel> fieldValues = this.findValueField(field, model.getProperties());
            for (org.opencdmp.commonmodels.models.description.FieldModel fieldValue : fieldValues) {
                if (fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
                    try {
                        value = Language.fromValue(fieldValue.getTextValue());
                        return value;
                    } catch (IllegalArgumentException | MyApplicationException e) {
                        logger.error(e.getMessage(), e);
                    }
                }
                if (fieldValue.getTextListValue() != null && !fieldValue.getTextListValue().isEmpty()) {
                    for (String val : fieldValue.getTextListValue()) {
                        try {
                            value = Language.fromValue(val);
                            return value;
                        } catch (IllegalArgumentException | MyApplicationException e) {
                            logger.error(e.getMessage(), e);
                        }
                    }
                }
            }
        }
        return buildRdaLanguage(model.getDescriptionTemplate().getLanguage());
    }

    private String buildRdaDatasetType(DescriptionModel model){
        for (FieldModel field : this.templateFieldSearcherService.searchFieldsBySemantics(model.getDescriptionTemplate(), SEMANTIC_DATASET_TYPE)) {
            List<org.opencdmp.commonmodels.models.description.FieldModel> fieldValues = this.findValueField(field, model.getProperties());
            for (org.opencdmp.commonmodels.models.description.FieldModel fieldValue : fieldValues) {
                if (fieldValue.getTextValue() != null && !fieldValue.getTextValue().isBlank()) {
                    return fieldValue.getTextValue();
                }

                if (fieldValue.getTextListValue() != null && !fieldValue.getTextListValue().isEmpty()) {
                    return fieldValue.getTextListValue().getFirst();
                }
            }
        }
        return "DMP Dataset";
    }

    private List<org.opencdmp.commonmodels.models.description.FieldModel> findValueField(FieldModel fieldModel, PropertyDefinitionModel descriptionTemplateModel){
        List<org.opencdmp.commonmodels.models.description.FieldModel> items = new ArrayList<>();
        if (fieldModel == null || descriptionTemplateModel == null || descriptionTemplateModel.getFieldSets() == null) return items;
        for (PropertyDefinitionFieldSetModel propertyDefinitionFieldSetModel : descriptionTemplateModel.getFieldSets().values()){
            if (propertyDefinitionFieldSetModel.getItems() == null) continue;
            for (PropertyDefinitionFieldSetItemModel propertyDefinitionFieldSetItemModel : propertyDefinitionFieldSetModel.getItems()){
                if (propertyDefinitionFieldSetItemModel.getFields() == null) continue;
                org.opencdmp.commonmodels.models.description.FieldModel valueField = propertyDefinitionFieldSetItemModel.getFields().getOrDefault(fieldModel.getId(), null);
                if (valueField != null) items.add(valueField);
            }
        }
        return items;
    }

    private List<org.opencdmp.commonmodels.models.description.PropertyDefinitionFieldSetItemModel> findFieldSetValue(FieldSetModel fieldSetModel, PropertyDefinitionModel descriptionTemplateModel){
        List<org.opencdmp.commonmodels.models.description.PropertyDefinitionFieldSetItemModel> items = new ArrayList<>();
        if (fieldSetModel == null || descriptionTemplateModel == null || descriptionTemplateModel.getFieldSets() == null) return items;
        PropertyDefinitionFieldSetModel propertyDefinitionFieldSetModel =  descriptionTemplateModel.getFieldSets().getOrDefault(fieldSetModel.getId(), null);
        if (propertyDefinitionFieldSetModel != null && propertyDefinitionFieldSetModel.getItems() != null) return propertyDefinitionFieldSetModel.getItems();
        return items;
    }
    
    private org.opencdmp.commonmodels.models.description.FieldModel findValueFieldBySemantic(FieldSetModel fieldSet, org.opencdmp.commonmodels.models.description.PropertyDefinitionFieldSetItemModel propertyDefinitionFieldSetItemModel, String semantic){
        FieldModel field = this.templateFieldSearcherService.findFieldBySemantic(fieldSet, semantic);
        return field != null ? propertyDefinitionFieldSetItemModel.getFields().getOrDefault(field.getId(), null) : null;
    }
    //endregion
    
    //endregion
    
    @Override
    public DmpModel importDmp(FileEnvelopeModel envelope) {
        throw new UnsupportedOperationException("import not supported");
    }

    @Override
    public DescriptionModel importDescription(FileEnvelopeModel envelope) {
        throw new UnsupportedOperationException("import not supported");
    }

    @Override
    public FileTransformerConfiguration getConfiguration() {
        List<FileFormat> supportedFormats = List.of(new FileFormat("json", false, null));
        FileTransformerConfiguration configuration = new FileTransformerConfiguration();
        configuration.setFileTransformerId(this.configuration.getRdaFileTransformerServiceProperties().getTransformerId());
        configuration.setExportVariants(supportedFormats);
        configuration.setExportEntityTypes(FILE_TRANSFORMER_ENTITY_TYPES);
        configuration.setImportVariants(null);
        configuration.setUseSharedStorage(this.configuration.getRdaFileTransformerServiceProperties().isUseSharedStorage());
        return configuration;
    }
}
